#include "netlinkthread.hh"

NetLinkThread::NetLinkThread(int nlgroup) 
{
  Init(nlgroup);
  incomming=true;
}

NetLinkThread::NetLinkThread(int nlgroup,bool direction)
{
  Init(nlgroup);
  incomming=direction;
}

void NetLinkThread::Init(int nlgroup)
{
  if ((nlgroup<1)||(nlgroup>32))
    throw string("Wrong netlink group. Must be 1-32!");
  h = ipulog_create_handle(ipulog_group2gmask(nlgroup));

  if (!h) {
    throw string("Error creating ulog handle");
  }
  buf = (unsigned char *) malloc(BUFFER_SIZE);

  running=true;
  incomming=true;
}

NetLinkThread::~NetLinkThread()
{
  ipulog_destroy_handle(h);
  delete(buf);
  running=false;
}

unsigned short NetLinkThread::SwapBytes(unsigned short to_swap) {
  return (((to_swap&0x00FF) << 8) | ((to_swap&0xFF00) >> 8));
}

void NetLinkThread::PaketHandler(ulog_packet_msg_t *pkt) {
  struct iphdr *iph = (struct iphdr *) pkt->payload;
  char ip[50];
  mutex.enterMutex();
  if (incomming)
    traffic[iph->daddr]+=SwapBytes(iph->tot_len);
  else
    traffic[iph->saddr]+=SwapBytes(iph->tot_len);
  mutex.leaveMutex();
}

map<unsigned int,unsigned long long> NetLinkThread::GetAndClear()
{

  char ip[50];
  mutex.enterMutex();  
  map<unsigned int, unsigned long long> copytraffic(traffic);
  traffic.clear();
  mutex.leaveMutex();
  return copytraffic;
}

void NetLinkThread::run()
{
  int len;
  while (running) {
    len = ipulog_read(h,buf,BUFFER_SIZE,1);
    if (len < 0 )
      cerr << "Short packet received" << endl;
    else
      while ((upkt = ipulog_get_packet(h,buf,len)))
	     PaketHandler(upkt);
  }
}
