#include "defines.h"
#include <netlinkthread.hh>
#include <string>
#include <iostream>
#include <vector>
#include <mysql/mysql.h>
#include <sys/syslog.h>
#include <signal.h>
#include <cc++/common.h>

#ifdef  CCXX_NAMESPACES
using namespace std;
using namespace ost;
#endif  

#define DIRECTION_IN true
#define DIRECTION_OUT false

struct MySQLServer {
  string hostname;
  string username;
  string password;
  string database;
  unsigned short port;
  bool connected;
  bool senddata;
  MYSQL *connection;
};

class UlogipacConfNetlink : public Keydata
{
  public:
    UlogipacConfNetlink() : Keydata("/ulogipac/netlink") {
    }
    unsigned int GetDirection(bool direction) {
      if (direction==DIRECTION_IN)
        return atoi(getFirst("netlink_group_in"));
      else
        return atoi(getFirst("netlink_group_out"));
      return 0;
    }
    int CheckConfig() {
      if (getCount("netlink_group_in")!=1 ||
          getCount("netlink_group_out")!=1 ) 
        return 1;
      else
        return 0; 
    }
};

class UlogipacConfMysql : public Keydata 
{
  public:
    UlogipacConfMysql() : Keydata("/ulogipac/mysql") {
    }

    int CheckConfig() {
      //FIXME: check for more errors
      unsigned short servercount;
      servercount=getCount("server");
      if (servercount==0) return -1;
      if (servercount!=getCount("username")) return 1;
      if (servercount!=getCount("password")) return 2;
      if (servercount!=getCount("database")) return 3;
      if (servercount!=getCount("port")) return 4;
      if (getCount("storage_interval")!=1) return 5;
      if (getCount("table_prefix_in")!=1) return 6;
      if (getCount("table_prefix_out")!=1) return 7;
      return 0; 
    }
    unsigned int GetStorageInterval() {
      return atoi(this->getFirst("storage_interval"));
    }

    const char *GetTablePrefix(bool direction) {
      if (direction==DIRECTION_IN) 
        return getFirst("table_prefix_in");
      else
        return getFirst("table_prefix_out");
    }
    const char *GetTableStructure(bool direction) {
      if (direction==DIRECTION_IN)
        return getFirst("create_table_in");
      else
        return getFirst("create_table_out");
    }
    const char *GetInsertQuery(bool direction) {
      if (direction==DIRECTION_IN)
        return getFirst("insert_in");
      else
        return getFirst("insert_out");
    }
    const char *GetUserIdQuery() {
        if (getCount("userid_query")>=1) 
	  return getFirst("userid_query");
        else
          return "no";
    }
    bool QueryUserId() {
        if (strcmp(getFirst("do_userid_query"),"yes")==0) 
          return true;
        else 
          return false;
    } 

    vector <MySQLServer> GetSQLServer() {
      vector <MySQLServer> retval;
      MySQLServer listentry;
      unsigned short count=getCount("server");
      const char * const *serverlist;
      const char * const *usernamelist;
      const char * const *passwordlist;
      const char * const *databaselist;
      const char * const *portlist;
      serverlist=this->getList("server");
      usernamelist=this->getList("username");
      passwordlist=this->getList("password");
      databaselist=this->getList("database");
      portlist=this->getList("port");
      for (unsigned int i=0;i<count;i++) {
        listentry.hostname=serverlist[i];
        listentry.username=usernamelist[i];
        listentry.password=passwordlist[i];
        listentry.database=databaselist[i];
        listentry.port=atoi(portlist[i]);
        listentry.connected=false;
	listentry.senddata=false;
        retval.push_back(listentry);
      }
      return retval;
    }
};

UlogipacConfMysql *mysqlconf;
vector <MySQLServer> mysqlserver;
bool direction,flushdata,running,query_userid;

CommandOptionNoArg helparg (
  "help", "?", "Print help usage"
);

CommandOptionArg directionarg (
  "direction", "d", "Set direction for trafficaccounting (in or out)",true
);

void CommandOptionHandling(int argc,char **argv) 
{
  CommandOptionParse *args = makeCommandOptionParse(
    argc,argv,
    "<-d direction>"
  );
  if ( helparg.numSet ) {
    cerr << args->printUsage();
    ::exit(0);
  }
  if ( args->argsHaveError() ) {
    cerr << args->printErrors();
    cerr << args->printUsage();
    ::exit(1);
  }
  if ( (strcmp(directionarg.values[0],"in")!=0) &&
       (strcmp(directionarg.values[0],"out")!=0) ) {
    cerr << "Unknown direction " << directionarg.values[0] << endl;
    cerr << "Valid values are: <in> <out>" << endl;
    ::exit(1);
  }
  if (strcmp(directionarg.values[0],"in")==0) direction=DIRECTION_IN;
  if (strcmp(directionarg.values[0],"out")==0) direction=DIRECTION_OUT;
}

void SignalHandler(int signal) 
{
  switch(signal) {
    case SIGUSR1: 
      flushdata=true;
      syslog(LOG_INFO,"Received flush data signal");
    break;
    case SIGTERM: 
      flushdata=true;
      running=false;
      syslog(LOG_INFO,"Received termination signal. Shutting down.");
    break;
    default: break;
  }
}

void ConnectSQL() 
{ 
  MYSQL *tempconn;
  char temp[500],ipstring[16];
  unsigned int counter=0;
  for (unsigned int i=0;i<mysqlserver.size();i++) {
    tempconn = new MYSQL;
    mysql_init(tempconn);
    sprintf(ipstring,"%s",mysqlserver[i].hostname.c_str());
    snprintf(temp,sizeof(temp),"Starting connection to MySQL Server %s",ipstring);
    syslog(LOG_INFO,temp);
    if (!mysql_real_connect(tempconn,
                            mysqlserver[i].hostname.c_str(),
                            mysqlserver[i].username.c_str(),
                            mysqlserver[i].password.c_str(),
                            mysqlserver[i].database.c_str(),
                            mysqlserver[i].port,
                            NULL,0)) {
      syslog(LOG_ERR,"Failed to connect to mysql database! Error: %s",mysql_error(tempconn));
      mysqlserver[i].connected=false;
      mysqlserver[i].senddata=false;
      mysqlserver[i].connection=NULL;
      delete(tempconn);
    } else {
      snprintf(temp,sizeof(temp),"Connection to MySQL Server %s established",ipstring);
      syslog(LOG_INFO,temp);
      mysqlserver[i].connected=true;
      mysqlserver[i].senddata=true;
      mysqlserver[i].connection=tempconn;
      counter++;
    }
  }
  sprintf(temp,"Connected to %i out of %i MySQL servers",counter,mysqlserver.size());
  syslog(LOG_INFO,temp);
}

void DisconnectSQL() 
{
  char temp[500];
  for (unsigned int i=0;i<mysqlserver.size();i++) {
    if (mysqlserver[i].connected) {
      mysql_close(mysqlserver[i].connection);
      mysqlserver[i].connected=false;
      sprintf(temp,"Closing connection to %s",mysqlserver[i].hostname.c_str());
      syslog(LOG_INFO,temp);
    }
  }
}

int main(int argc, char **argv) 
{
  char syslogst[500];
  flushdata=false;running=true;query_userid=false;

  CommandOptionHandling(argc,argv);

  openlog(argv[0],LOG_PID,LOG_LOCAL6);
  syslog(LOG_DEBUG,"ulogipac starting");
  syslog(LOG_DEBUG,"Reading config from /etc/ulogipac.conf");

  /* check config */
  mysqlconf = new UlogipacConfMysql();
  int configcheck=mysqlconf->CheckConfig();
  if (configcheck!=0) {
    switch (configcheck) {
      // FIXME: add all errors
      case -1: syslog(LOG_ERR,"Config file error: No mysql server specified");
               break;
      case  1: syslog(LOG_ERR,"Config file error: Count of username does not match servers");
               break;
      case  2: syslog(LOG_ERR,"Config file error: Count of password does not match servers");
               break;
      case  3: syslog(LOG_ERR,"Config file error: Count of database does not match servers");
               break;
      case  4: syslog(LOG_ERR,"Config file error: Count of port does not match servers");
               break;

    }; 
    exit(2);
  }
  query_userid=mysqlconf->QueryUserId();
#ifdef DEBUG_ULOGIPAC
  cout << "QUERY userid " << query_userid << endl;
#endif

  mysqlserver=mysqlconf->GetSQLServer();

  
  UlogipacConfNetlink *netlinkconf = new UlogipacConfNetlink();

  if (netlinkconf->CheckConfig()!=0) {
    syslog(LOG_ERR,"Config file error: netlink_group_in or netlink_group_out not specified correctly");
    exit(3);
  }
#ifndef DEBUG_ULOGIPAC
  if (fork()==0) {
#endif
    signal(SIGUSR1,SignalHandler);
    signal(SIGTERM,SignalHandler);
    ConnectSQL();
    NetLinkThread *nlthread = new NetLinkThread(netlinkconf->GetDirection(direction),direction);
    try {
      nlthread->start();
    }
    catch (string st) {
      syslog(LOG_ERR,"Netlink thread startup error: %s",st.c_str());
    }

    char ip[20];
    char tablename[500],query[500],queryid[500],userid[100];
    map<unsigned int,unsigned long long> traffic;
    map<unsigned int,unsigned long long>::iterator it;
    time_t zeit_t;
    struct tm *zeit;
    char mypid[10];
    snprintf(mypid,sizeof(mypid),"%i",getpid());

    /* go to main loop */
    while (running) {
      /* wait until it is time to store data */
      for (unsigned int i=0;i<mysqlconf->GetStorageInterval();i++) {
        ::sleep(1);
        if (flushdata) {
          flushdata=false;
          break;
        }
      }
      /* check whether SQL server are online */
      int mysqlsendcounter=0;
      for (unsigned int i=0;i<mysqlserver.size();i++) {
        if (mysql_ping(mysqlserver[i].connection)!=0)
          mysqlserver[i].senddata=false;
        else {
          mysqlserver[i].senddata=true;
          mysqlsendcounter++;
        }
      }

      /* cannot write any more data to SQL servers */
      if (mysqlsendcounter<=0) {
        syslog(LOG_ERR,"SQL Server error. No more servers avaiable. Exiting!!!");
        exit(10);
      } 

      /* get current time and date for prefix of tables */
      zeit_t=time((time_t *)NULL);zeit=localtime(&zeit_t);
      snprintf(tablename,sizeof(tablename),"%s%02i_%02i",mysqlconf->GetTablePrefix(direction),zeit->tm_year+1900,zeit->tm_mon+1);

      // check tables
      snprintf(query,sizeof(query),"create table if not exists %s %s;",tablename,mysqlconf->GetTableStructure(direction));
      for (unsigned int i=0;i<mysqlserver.size();i++) {
        if (mysqlserver[i].senddata) 
          if (mysql_query(mysqlserver[i].connection,query)!=0) {
            // FIXME add hostname:port to error
            syslog(LOG_ERR,"error sending create table query to server! Stopping data stream to server!");
            mysqlserver[i].senddata=false;
          }
      }

      traffic=nlthread->GetAndClear();
      it=traffic.begin();
      while (it!=traffic.end()) {
        string selectstring,qstring,qstringleft,qstringright;
        char bytesstr[50],IPstr[50];
        snprintf(bytesstr,sizeof(bytesstr),"%lu",traffic[(*it).first]);
        sprintf(IPstr,"\"%u.%u.%u.%u\"",NIPQUAD((*it).first));

	snprintf(query,sizeof(query),"insert into %s %s",tablename,mysqlconf->GetInsertQuery(direction));
	qstring=string(query);

        if (query_userid) { 
          string qidstring,qidstringleft,qidstringright; 
          bzero(userid,sizeof(userid));
          snprintf(queryid,sizeof(queryid),"select %s",mysqlconf->GetUserIdQuery());
          qidstring=string(queryid);
          qidstringleft.assign(qidstring,0,qidstring.find("ULOGIPAC_IP"));
          qidstringright.assign(qidstring,qidstring.find("ULOGIPAC_IP")+string("ULOGIPAC_IP").size(),qidstring.size());
          qidstring=qidstringleft+string(IPstr)+qidstringright;
          int gotid=0;
	  for (unsigned int i=0;i<mysqlserver.size();i++) {
            if (mysqlserver[i].senddata) {
      	      if (mysql_query(mysqlserver[i].connection,qidstring.c_str())!=0) {
	        bzero(syslogst,sizeof(syslogst));
                snprintf(syslogst,sizeof(syslogst),"Error sending SQL SELECT query for userid. Query: %s",qidstring.c_str());
                syslog(LOG_ERR,syslogst);
#ifdef DEBUG_ULOGIPAC
		cout << "Error in: " << qidstring.c_str() << endl;
#endif
              } else {
		MYSQL_RES *result=NULL;
                result=mysql_store_result(mysqlserver[i].connection);
		if (result==NULL) {
                  bzero(syslogst,sizeof(syslogst));
	          snprintf(syslogst,sizeof(syslogst),"UserID Query returned NULL result. Query: %s",qidstring.c_str());
		  syslog(LOG_ERR,syslogst);
                } else {
		  if (mysql_num_rows(result)!=1) {
                    bzero(syslogst,sizeof(syslogst));
                    snprintf(syslogst,sizeof(syslogst),"UserID Query returned !=1 rows. Query: %s",qidstring.c_str());
                    syslog(LOG_ERR,syslogst);
                  } else {
                    if (mysql_num_fields(result) == 0) {
                      bzero(syslogst,sizeof(syslogst));
                      snprintf(syslogst,sizeof(syslogst),"UserID Query did not return valid data. Query: %s",qidstring.c_str());
                      syslog(LOG_ERR,syslogst);
                    } else {
                      MYSQL_ROW row;
		      row = mysql_fetch_row(result);
#ifdef DEBUG_ULOGIPAC
                      cout << IPstr << " " << row[0] << endl;
#endif
		      bzero(userid,sizeof(userid));
		      snprintf(userid,sizeof(userid),"\"%s\"",row[0]);
		      gotid=1;
                    }
                  }
 		  mysql_free_result(result);
	 	} /* end if result == NULL */
              }
            }
          } /* end for */
          if (gotid==0) 
            snprintf(userid,sizeof(userid),"NULL");
          qstringleft.assign(qstring,0,qstring.find("ULOGIPAC_USERID"));
          qstringright.assign(qstring,qstring.find("ULOGIPAC_USERID")+string("ULOGIPAC_USERID").size(),qstring.size());
          qstring=qstringleft+string(userid)+qstringright;
        } /* end if query_userid */

        qstringleft.assign(qstring,0,qstring.find("ULOGIPAC_PID"));
        qstringright.assign(qstring,qstring.find("ULOGIPAC_PID")+string("ULOGIPAC_PID").size(),qstring.size());
        qstring=qstringleft+string(mypid)+qstringright;

        qstringleft.assign(qstring,0,qstring.find("ULOGIPAC_IP"));
        qstringright.assign(qstring,qstring.find("ULOGIPAC_IP")+string("ULOGIPAC_IP").size(),qstring.size());
        qstring=qstringleft+string(IPstr)+qstringright;

        qstringleft.assign(qstring,0,qstring.find("ULOGIPAC_BYTES"));
        qstringright.assign(qstring,qstring.find("ULOGIPAC_BYTES")+string("ULOGIPAC_BYTES").size(),qstring.size());
        qstring=qstringleft+string(bytesstr)+qstringright;
#ifdef DEBUG_ULOGIPAC
        cout << "INSERT query: " << qstring.c_str() << endl;
#endif
        for (unsigned int i=0;i<mysqlserver.size();i++) {
          if (mysqlserver[i].senddata)
            if(mysql_query(mysqlserver[i].connection,qstring.c_str())!=0) {
	      bzero(syslogst,sizeof(syslogst));
              snprintf(syslogst,sizeof(syslogst),"Error sending SQL INSERT query. Query: %s",qstring.c_str()); 
              syslog(LOG_ERR,syslogst); 
              mysqlserver[i].senddata=false;
            }
        }
        it++;
      }
    } 
    DisconnectSQL();
#ifndef DEBUG_ULOGIPAC
  } /* end fork */
#endif
  
  closelog();
  return 0;
}
